from ratio.router.route import RouteMethod
from ratio.router.resolvers.resolver import ResolvedRoute
from ratio.http.response import Response
from ratio.http.request import Request


def wrap_bound_method(bound_method: RouteMethod, request: Request) -> ResolvedRoute:
    async def wrapped_method() -> Response:
        return await bound_method(request)

    return wrapped_method



from ratio.router.route_path import RoutePath
from ratio.router.route import Route
from ratio.http.response import Response
from ratio.http.request import Request
from abc import ABC, abstractmethod
from typing import Type, Callable, Awaitable

ResolvedRoute = Callable[[], Awaitable[Response]]


class RouteResolver(ABC):
    _collection: dict[str, Type[Route]]

    def __init__(self) -> None:
        self._collection = {}

    @abstractmethod
    def resolve(self, request: Request) -> ResolvedRoute | None:
        ...

    @abstractmethod
    def register(self, route_path: RoutePath, route: Type[Route]) -> None:
        ...



from ratio.router.resolvers.resolver import RouteResolver, ResolvedRoute
from ratio.router.exceptions import PathAlreadyExistsError
from ratio.router.resolvers.util import wrap_bound_method
from ratio.router.resolvers.generic_routes import method_not_allowed
from ratio.router.route_path import RoutePath
from ratio.router.route import Route
from ratio.http.request import Request
from typing import Type
import re
import copy


class DynamicResolver(RouteResolver):
    _collection: dict[str, Type[Route]]
    __registered_regex_formats: list[str]

    def __init__(self) -> None:
        super(DynamicResolver, self).__init__()
        self.__registered_regex_formats = []

    def resolve(self, request: Request) -> ResolvedRoute | None:
        matches = set(
            filter(
                lambda result: result is not None,
                map(
                    lambda regex: re.fullmatch(regex, request.url_path),
                    self._collection.keys(),
                ),
            )
        )

        if len(matches) == 0:
            return None

        if len(matches) > 1:
            return None

        if (matched_route := next(iter(matches))) is None:
            return None

        matched_regex = str(matched_route.re.pattern)
        route = self._collection.get(matched_regex, None)

        if route is None:
            return None

        instantiated_route = route()
        bound_method = getattr(instantiated_route, request.method.lower(), None)
        if bound_method is None:
            return method_not_allowed

        # We need to make a copy of the request, so we can set the path parameters.
        new_request = copy.deepcopy(request)
        new_request.path_parameters = matched_route.groupdict()
        return wrap_bound_method(bound_method, new_request)

    def register(self, route_path: RoutePath, route: Type[Route]) -> None:
        simplified_regex = route_path.as_simple_regex()

        if simplified_regex in self.__registered_regex_formats:
            raise PathAlreadyExistsError(route_path.path, simplified_regex)

        regex = route_path.as_regex()

        # To avoid we have to do this again, we save by regex, without the group ID:
        self.__registered_regex_formats.append(simplified_regex)
        self._collection[regex] = route



from ratio.router.router import Router
from ratio.router.route import Route
from ratio.router.route_path import RoutePath
from ratio.http.request import Request
from ratio.http.response import Response
from ratio.application.discover_routes import discover_routes
from asgiref.typing import Scope
from http import HTTPMethod, HTTPStatus
import pathlib


class Ratio:
    application_root: pathlib.Path
    router: Router

    def __init__(self, root: pathlib.Path | None):
        self.application_root = root if root is not None else pathlib.Path(__file__)

        routes = discover_routes(self.application_root)
        self.router = Router(routes)

    async def __call__(self, scope: Scope, receive, send):
        url = scope.get("path")
        request = Request(HTTPMethod.GET, url)
        route = await self.router.resolve(request)
        response = await route()

        await send(
            {
                "type": "http.response.start",
                "status": response.code,
                "headers": [
                    [b"content-type", b"text/plain"],
                ],
            }
        )
        await send(
            {
                "type": "http.response.body",
                "body": str.encode(response.message),
            }
        )


app = Ratio(pathlib.Path.cwd())


class TestRoute(Route):
    async def get(self, request: Request) -> Response:
        return Response.from_http_status(HTTPStatus.NOT_ACCEPTABLE)


app.router.register_route(RoutePath("/"), TestRoute)
app.router.register_route(RoutePath("/test/[id]"), Route)



import pathlib
import importlib.util
from types import ModuleType


def load_local_module(name: str, path: pathlib.Path) -> ModuleType:
    spec = importlib.util.spec_from_file_location(name, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module


def convert_snake_case_to_pascal_case(snaked_string: str) -> str:
    components = snaked_string.split("_")
    return "".join([component.capitalize() for component in components])



from abc import ABC
from typing import Type, TypeVar, Generic
from ratio.database.models.model import Model


T = TypeVar("T", bound=Model)


class Queryable(ABC, Generic[T]):
    __table: str
    __model: Type[T]
    __primary_key: str

    def __init__(self, model: Type[T]):
        self.__table = model.__class__.__name__.lower()
        self.__model = model

    def get_by_id(self):
        if not self.__primary_key == "id":
            raise NotImplementedError


class UserQueryable(Queryable):
    pass



from ratio.router.route import Route
from ratio.router.route_type import RouteType
from ratio.router.resolvers.resolver import ResolvedRoute
from ratio.router.resolvers.direct_resolver import DirectResolver
from ratio.router.resolvers.dynamic_resolver import DynamicResolver
from ratio.router.route_path import RoutePath
from ratio.http.request import Request
from typing import Type


class Router:
    __dynamic_resolver: DynamicResolver
    __direct_resolver: DirectResolver

    def __init__(self, routes: dict[RoutePath, Type[Route]]) -> None:
        self.__dynamic_resolver = DynamicResolver()
        self.__direct_resolver = DirectResolver()

        for path, route in routes.items():
            self.register_route(path, route)

    async def resolve(self, request: Request) -> ResolvedRoute:
        # First try direct route resolution
        if (route := self.__direct_resolver.resolve(request)) is not None:
            # Since this is an ABC, we know for sure these methods are implemented.
            # So there is no reason to take any additional precautions. It's fine to
            # throw an error when an HTTP method does not exist on the Route class.
            return route

        # Try dynamic resolution for this route
        if (dynamic_route := self.__dynamic_resolver.resolve(request)) is not None:
            return dynamic_route

        return self.__direct_resolver.not_found_route(request)

    def register_route(self, path: RoutePath, route: Type[Route]) -> None:
        # No dynamic parameters means that a path is considered direct
        if path.route_type == RouteType.DIRECT:
            return self.__direct_resolver.register(path, route)

        return self.__dynamic_resolver.register(path, route)



import argparse


def _create_argument_parser() -> argparse.ArgumentParser:
    main_parser = argparse.ArgumentParser(
        conflict_handler="resolve",
        allow_abbrev=False,
        add_help=False,
        exit_on_error=False,
    )
    main_parser.add_argument("command")

    return main_parser



from ratio.http.response import Response
from ratio.http.request import Request
from typing import Awaitable, Callable
from http import HTTPStatus

RouteMethod = Callable[[Request], Awaitable[Response]]


class Route:
    async def get(self, _request: Request) -> Response:
        return Response.from_http_status(HTTPStatus.METHOD_NOT_ALLOWED)

    async def head(self, _request: Request) -> Response:
        return Response.from_http_status(HTTPStatus.METHOD_NOT_ALLOWED)

    async def patch(self, _request: Request) -> Response:
        return Response.from_http_status(HTTPStatus.METHOD_NOT_ALLOWED)

    async def post(self, _request: Request) -> Response:
        return Response.from_http_status(HTTPStatus.METHOD_NOT_ALLOWED)

    async def put(self, _request: Request) -> Response:
        return Response.from_http_status(HTTPStatus.METHOD_NOT_ALLOWED)

    async def delete(self, _request: Request) -> Response:
        return Response.from_http_status(HTTPStatus.METHOD_NOT_ALLOWED)

    async def trace(self, _request: Request) -> Response:
        return Response.from_http_status(HTTPStatus.METHOD_NOT_ALLOWED)

    async def options(self, _request: Request) -> Response:
        return Response.from_http_status(HTTPStatus.METHOD_NOT_ALLOWED)

    async def connect(self, _request: Request) -> Response:
        return Response.from_http_status(HTTPStatus.METHOD_NOT_ALLOWED)



class InvalidRouteError(ValueError):
    pass


class PathAlreadyExistsError(ValueError):
    def __init__(self, path: str, regex: str | None = None):
        if regex is None:
            self.message = f"The following direct path already exists: '{path}'"
        else:
            self.message = f"The following dynamic path already exists, because the path '{path}' signature '{regex}' already exists.'"

        super().__init__(self.message)



from ratio.router.route_type import RouteType
from ratio.router.exceptions import InvalidRouteError
import re
from functools import cached_property


ROUTE_ALLOWED_CHARACTERS = r"^[A-Za-z0-9\_\/\[\]]+$"
DYNAMIC_ROUTE_ALLOWED_CHARACTERS = r"[A-Za-z0-9\-_]+"
DYNAMIC_PATH_PARAMETER_REGEX = r"/\[[A-Za-z0-9]+\]"
INVALID_DYNAMIC_PARAMETER_REGEX = r"(.*?)\[(.*?)\/(.*?)](.*?)$"


class RoutePath:
    path: str

    def __init__(self, path: str):
        if path == "/":
            path = "/index"

        self.path = path
        self.assert_is_valid()

    @property
    def route_type(self) -> RouteType:
        if len(self.parameter_keys) == 0:
            return RouteType.DIRECT

        return RouteType.DYNAMIC

    @cached_property
    def is_valid(self) -> bool:
        try:
            self.assert_is_valid()
            return True
        except InvalidRouteError:
            return False

    def assert_is_valid(self) -> None:
        if not self.path.startswith("/"):
            raise InvalidRouteError(
                "A route path should start with `/`, indicating the project root."
            )

        if self.path.endswith("/"):
            raise InvalidRouteError(
                "A route path must refer to a file, not a directory. It may not end with `/`."
            )

        if not re.match(ROUTE_ALLOWED_CHARACTERS, self.path):
            raise InvalidRouteError(
                "Routes may only contain underscores, alphanumerical characters and `/`."
            )

        if self.route_type == RouteType.DIRECT:
            return

        if not self.path.count("[") == self.path.count("]") or self.path.find(
            "["
        ) > self.path.find("]"):
            # TODO: catch /test/[id]/]id/whatever[
            raise InvalidRouteError(
                "A route with dynamic parameters should not leave open brackets."
            )

        if re.match(INVALID_DYNAMIC_PARAMETER_REGEX, self.path) is not None:
            # TODO: catch /[te/i] /[te[id]]
            raise InvalidRouteError(
                "A dynamic parameter name may not contain forward slashes."
            )

        # If a dynamic parameter key appears twice, a route is invalid
        if len(set(self.parameter_keys)) != len(self.parameter_keys):
            raise InvalidRouteError(
                "A dynamic route should contain unique parameter keys."
            )

        # For custom error routes to work, we cannot have `/[id]` as route.
        # TODO: is this really the case cause direct routes have precedence over dynamic routes
        # On the other hand, `/[id]/something` is fine.
        if len(self.parameter_keys) == 1 and len(self.path.split("/")) < 3:
            raise InvalidRouteError("Top-level dynamic routes are not allowed.")

    @cached_property
    def parameter_keys(self) -> list[str]:
        keys = re.findall(DYNAMIC_PATH_PARAMETER_REGEX, self.path)
        return [key[2:][:-1] for key in keys]

    def as_simple_regex(self) -> str:
        if self.route_type == RouteType.DIRECT:
            return self.path

        regex = self.path

        for parameter in self.parameter_keys:
            regex = re.sub(
                rf"/\[{parameter}\]", f"/({DYNAMIC_ROUTE_ALLOWED_CHARACTERS})", regex
            )

        return regex

    def as_regex(self) -> str:
        if self.route_type == RouteType.DIRECT:
            return self.path

        regex = self.path

        for parameter in self.parameter_keys:
            regex = re.sub(
                rf"/\[{parameter}\]",
                f"/(?P<{parameter}>{DYNAMIC_ROUTE_ALLOWED_CHARACTERS})",
                regex,
            )

        return regex



from ratio.router.resolvers.resolver import RouteResolver, ResolvedRoute
from ratio.router.resolvers.generic_routes import not_found, method_not_allowed
from ratio.router.resolvers.util import wrap_bound_method
from ratio.router.exceptions import PathAlreadyExistsError
from ratio.router.route_path import RoutePath
from ratio.router.route import Route
from ratio.http.request import Request
from typing import Type


class DirectResolver(RouteResolver):
    def resolve(self, request: Request) -> ResolvedRoute | None:
        route = self._collection.get(request.url_path, None)

        if route is None:
            return None

        instantiated_route = route()
        bound_method = getattr(instantiated_route, request.method.lower(), None)
        if bound_method is None:
            return method_not_allowed

        return wrap_bound_method(bound_method, request)

    def register(self, route_path: RoutePath, route: Type[Route]) -> None:
        if route_path.path in self._collection:
            raise PathAlreadyExistsError(route_path.path)

        self._collection[route_path.path] = route

    def not_found_route(self, request: Request) -> ResolvedRoute:
        route = self._collection.get("/404", None)

        if route is None:
            return not_found

        instantiated_route = route()
        bound_method = getattr(instantiated_route, request.method.lower(), None)

        if bound_method is None:
            return method_not_allowed

        return wrap_bound_method(bound_method, request)



from ratio.http.response import Response
from ratio.router.resolvers.resolver import ResolvedRoute
from http import HTTPStatus


def _create_generic_route(status: HTTPStatus) -> ResolvedRoute:
    async def generic_route() -> Response:
        return Response.from_http_status(status)

    return generic_route


not_found: ResolvedRoute = _create_generic_route(HTTPStatus.NOT_FOUND)
internal_server_error: ResolvedRoute = _create_generic_route(
    HTTPStatus.INTERNAL_SERVER_ERROR
)
method_not_allowed: ResolvedRoute = _create_generic_route(HTTPStatus.METHOD_NOT_ALLOWED)



from ratio.http.parameters import Parameters
from http import HTTPMethod
import urllib.parse


class Request:
    method: HTTPMethod
    url: str
    body: list[str]
    cookies: dict[str, str]
    fresh: bool
    path_parameters: dict[str, str]

    __parsed_url: urllib.parse.ParseResult

    def __init__(self, method: HTTPMethod, url: str) -> None:
        self.__parsed_url = urllib.parse.urlparse(url)
        self.method = method
        self.path_parameters = {}

    @property
    def url_hostname(self) -> str | None:
        return self.__parsed_url.hostname

    @property
    def url_port(self) -> int | None:
        return self.__parsed_url.port

    @property
    def url_netloc(self) -> str:
        return self.__parsed_url.netloc

    @property
    def url_path(self) -> str:
        if self.__parsed_url.path == "/":
            return "/index"

        return self.__parsed_url.path

    @property
    def protocol(self) -> str:
        return self.__parsed_url.scheme

    @property
    def query_parameters(self) -> dict[str, str | list[str]]:
        parsed_query_parameters = urllib.parse.parse_qs(self.__parsed_url.query)

        return {
            key: value[0] if len(value) == 1 else value
            for (key, value) in parsed_query_parameters.items()
        }

    @property
    def parameters(self) -> Parameters:
        return {"query": self.query_parameters, "path": self.path_parameters}



from ratio.console._create_argument_parser import _create_argument_parser
from ratio.console._create_rich_console import _create_rich_console
from ratio.application.application import Application


class ConsoleApplication(Application):
    def __init__(self) -> None:
        super().__init__()


def run_console_application(arguments: list[str]) -> None:
    parser = _create_argument_parser()
    parsed = parser.parse_args(arguments)

    if parsed.command not in []:
        raise NotImplementedError

    # Set up application environment
    console = _create_rich_console()
    console.print("[bold]ratio {command}[/bold]")



from ratio.application.util import load_local_module, convert_snake_case_to_pascal_case
from ratio.router.route import Route
from ratio.router.route_path import RoutePath
import pathlib
from typing import Type
from enum import Enum


class RouteAction(Enum):
    ADDED = "ADDED"
    MISSING_ROUTE_CLASS = "MISSING_ROUTE_CLASS"


def guess_route_class_name(path: pathlib.Path):
    stem = path.stem
    if not stem.endswith("_route"):
        stem += "_route"

    return convert_snake_case_to_pascal_case(stem)


def map_discovered_route_to_route(
    path: pathlib.Path,
) -> tuple[Type[Route], RouteAction]:
    resolved_path = path.resolve()
    expected_class_name = guess_route_class_name(resolved_path)

    route = load_local_module("route", resolved_path)

    if not hasattr(route, expected_class_name):
        return resolved_path, RouteAction.MISSING_ROUTE_CLASS

    return resolved_path, "added"


def discover_routes(application_root: pathlib.Path) -> dict[RoutePath, Type[Route]]:
    routes_directory = application_root / "routes"
    if not routes_directory.exists() or not routes_directory.is_dir():
        return {}

    potential_routes = map(
        map_discovered_route_to_route, routes_directory.glob("**/*.py")
    )

    return {}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ratio',
 'ratio.application',
 'ratio.console',
 'ratio.database',
 'ratio.database.fields',
 'ratio.database.migrations',
 'ratio.database.models',
 'ratio.database.queries',
 'ratio.database.seeders',
 'ratio.http',
 'ratio.router',
 'ratio.router.resolvers']

package_data = \
{'': ['*']}

install_requires = \
['rich>=12.6.0,<13.0.0']

setup_kwargs = {
    'name': 'ratio',
    'version': '0.1.1',
    'description': 'The Python Web Framework for developers who like to get shit done',
    'long_description': '<h1 align="center">Ratio</h1>\n<p align="center">\n  The Python web framework for developers who want to get shit done.\n</p>\n<hr />\n<strong>Ratio is currently being developed and all releases in this phase may introduce breaking changes until further notice.\nPlease do not use Ratio in production without carefully considering the consequences.</strong>\n<hr />\n<h2>What is Ratio?</h2>\n<p>\n  Ratio is an asynchronous Python web framework that was built with developer experience in mind. Quick to learn for those\n  who just start out with programming and powerful so that senior developers can build high performance web applications \n  with it. The framework is designed with the Goldilocks principle in mind: just enough. Enough power to run high performance\n  web applications, enough intuitive design, so that developers can easily pick up on the principles.\n</p>\n<p>\n  Ratio borrows ideas from great frameworks, like <a href="https://github.com/django/django" target="_blank">Django</a>, \n  <a href="https://github.com/tiangolo/fastapi" target="_blank">FastAPI</a> and <a href="https://github.com/vercel/next.js" target="_blank">Next.js</a>, \n  and combines them with original concepts to improve the life of a developer when creating web applications for any\n  purpose.\n</p>\n<h2>Ready out of the box</h2>\n<p>\n  Ratio will be shipped with a custom and extensible command line interface, which can be used to perform actions within a\n  project.\n</p>\n<p>\n  This is what we aim Ratio to do:<br>\n  <small>This list is not complete and will be extended after certain releases in the pre-release phase.</small>\n</p>\n\n<ul>\n  <li><strong>File based routing:</strong> Intuitive routing for each incoming request, based on file system.</li>\n  <li><strong>Integrates with databases:</strong> Connect to SQL or SQLite databases from within the application controllers.</li>\n  <li><strong>Write once, use everywhere:</strong> Do not repeat yourself, by defining models, routes and actions you can use them throughout the application.</li>\n  <li><strong>Adheres to standards:</strong> API views are based on <a href="">OpenAPI</a> and the JSON schema standard.</li>\n</ul>\n\n\n<h2>Minimal external dependencies</h2>\n<p>\n  Currently, Ratio only requires the <code>rich</code> package from outside the Python standard library, which is used \n  for rendering beautiful output to the command line. In a future version of Ratio, we might want to remove this direct\n  dependency for users who really want to have no external dependencies.\n</p>',
    'author': 'Job Veldhuis',
    'author_email': 'job@baukefrederik.me',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)



from http import HTTPStatus
from typing import Self, TypedDict


class ResponseData(TypedDict):
    code: int
    message: str


class Response:
    """Generic response class for all default responses"""

    code: int
    message: str
    __success: bool

    def __init__(self, data: ResponseData) -> None:
        code = data["code"]
        message = data["message"]

        try:
            assert 100 < code < 600
        except AssertionError:
            code = 500

        self.code = code
        self.message = message

    # Reason for ignoring type: Self is actually valid. In new version of Mypy this will be supported.
    @classmethod
    def from_http_status(cls, status: HTTPStatus) -> Self:  # type: ignore
        return cls({"code": status.value, "message": status.phrase})

    @property
    def success(self) -> bool:
        return 200 <= self.code <= 299
